from __future__ import unicode_literals

import json

from mock import patch

from core import utils
from django.test import TestCase
from core.models import ElectionResults
from .utils import ScrutinyTestUtils


class TestScrutinyFunctions(ScrutinyTestUtils, TestCase):
    """ """
    def setUp(self):
        """Setup all needed endpoints."""
        self.election_id = 5
        # Simulated data
        self.received_data = self.create_received_data()
        # Vote objects from the AS
        self.vote_objs = []
        for i in range(0, 5):
            self.vote_objs.append(self.create_vote_obj(id=i))

        # List of deciphered votes
        self.deciphered_votes = self.create_vote_batch()

    # is_worker_running unit tests

    def test_is_worker_running_false(self):
        worker_status = utils.is_worker_running(pid=1324)
        self.assertFalse(worker_status)

    @patch("core.utils.os.kill")
    def test_is_worker_running_true(self, mock_fn):
        mock_fn.return_value = True
        worker_status = utils.is_worker_running(pid=1324)
        self.assertTrue(worker_status)

    # fuse_shared_keys unit tests

    def test_fuse_shared_keys_empty(self):
        with self.assertRaises(Exception):
            utils.fuse_shared_keys([], 1, "")

    @patch("core.utils.shamirsecret.recoverSecretFromComponents")
    def test_fuse_shared_keys_incorrect(self, mock_fn):
        mock_fn.return_value = (1, None)
        with self.assertRaises(Exception):
            utils.fuse_shared_keys(["1", "2", "3"], 1, "dummy")

    @patch("core.utils.shamirsecret.recoverSecretFromComponents")
    def test_fuse_shared_keys_success(self, mock_fn):
        secret = "dummysecret"
        mock_fn.return_value = (None, secret)
        shared_secret = utils.fuse_shared_keys(["1", "2", "3"], 1, "dummy")
        self.assertEquals(secret, shared_secret)

    # async_scrutinize unit tests

    def test_async_scrutinize_error(self):
        results_obj = ElectionResults(election_id=self.election_id)
        results_obj.save()
        utils.async_scrutinize(results_obj, {})
        self.assertEquals(
            results_obj.result_status_choices,
            ElectionResults.ResultStatus.ERROR
        )

    @patch("core.utils.routers.ASrouter")
    @patch("core.utils.fuse_shared_keys")
    def test_async_scrutinize_invalid_signatures(self, mock_fuse, mock_router):
        mock_fuse.return_value = "dummysecret"
        mock_router.return_value.fetch_votes.return_value = \
            (self.vote_objs, False)
        results_obj = ElectionResults(election_id=self.election_id)
        results_obj.save()
        utils.async_scrutinize(results_obj, self.received_data)
        results = json.loads(results_obj.election_results)
        # Results were properly published
        self.assertEquals(
            results_obj.result_status_choices,
            ElectionResults.ResultStatus.PUBLISHED
        )
        # All votes were counted
        self.assertEquals(results['total_votes'], len(self.vote_objs))
        # All votes marked as invalid
        self.assertEquals(results['invalid_votes'], len(self.vote_objs))
        # Vote invalidation by signature
        self.assertEquals(results['invalid_signature'], len(self.vote_objs))

    @patch("core.utils.eccblind.verifySignature")
    @patch("core.utils.routers.ASrouter")
    @patch("core.utils.fuse_shared_keys")
    def test_async_scrutinize_invalid_cipher(self, mock_fuse, mock_router,
                                             mock_verifysig):
        mock_verifysig.return_value = (None, True)
        mock_fuse.return_value = "dummysecret"
        mock_router.return_value.fetch_votes.return_value = \
            (self.vote_objs, False)
        results_obj = ElectionResults(election_id=self.election_id)
        results_obj.save()
        utils.async_scrutinize(results_obj, self.received_data)
        results = json.loads(results_obj.election_results)
        # Results were properly published
        self.assertEquals(
            results_obj.result_status_choices,
            ElectionResults.ResultStatus.PUBLISHED
        )
        # All votes were counted
        self.assertEquals(results['total_votes'], len(self.vote_objs))
        # All votes marked as invalid
        self.assertEquals(results['invalid_votes'], len(self.vote_objs))
        # Vote invalidation by signature
        self.assertEquals(results['invalid_cipher'], len(self.vote_objs))

    @patch("core.utils.pkiutils.decryptWithPrivateKey")
    @patch("core.utils.eccblind.verifySignature")
    @patch("core.utils.routers.ASrouter")
    @patch("core.utils.fuse_shared_keys")
    def test_async_scrutinize_cs_sign(self, mock_fuse, mock_router,
                                      mock_verifysig, mock_decrypt):
        mock_decrypt.return_value = (None, json.dumps(self.deciphered_votes))
        mock_verifysig.return_value = (None, True)
        mock_fuse.return_value = "dummysecret"
        mock_router.return_value.fetch_votes.return_value = \
            (self.vote_objs, False)
        results_obj = ElectionResults(election_id=self.election_id)
        results_obj.save()
        utils.async_scrutinize(results_obj, self.received_data)
        self.assertEquals(
            results_obj.result_status_choices,
            ElectionResults.ResultStatus.ERROR
        )

    @patch("core.utils.pkiutils.signObject")
    @patch("core.utils.pkiutils.decryptWithPrivateKey")
    @patch("core.utils.eccblind.verifySignature")
    @patch("core.utils.routers.ASrouter")
    @patch("core.utils.fuse_shared_keys")
    def test_async_scrutinize_failed_batch(self, mock_fuse, mock_router,
                                           mock_verifysig, mock_decrypt,
                                           mock_sign_obj):
        mock_sign_obj.return_value = (None, "signedvote")
        mock_decrypt.return_value = (None, json.dumps(self.deciphered_votes))
        mock_verifysig.return_value = (None, True)
        mock_fuse.return_value = "dummysecret"
        mock_router.return_value.fetch_votes.return_value = \
            (self.vote_objs, False)
        mock_router.return_value.sign_votes.return_value = \
            None
        results_obj = ElectionResults(election_id=self.election_id)
        results_obj.save()
        utils.async_scrutinize(results_obj, self.received_data)
        # Results were properly published
        self.assertEquals(
            results_obj.result_status_choices,
            ElectionResults.ResultStatus.ERROR
        )

    @patch("core.utils.pkiutils.signObject")
    @patch("core.utils.pkiutils.decryptWithPrivateKey")
    @patch("core.utils.eccblind.verifySignature")
    @patch("core.utils.routers.ASrouter")
    @patch("core.utils.fuse_shared_keys")
    def test_async_scrutinize_failed_update(self, mock_fuse, mock_router,
                                            mock_verifysig, mock_decrypt,
                                            mock_sign_obj):
        mock_sign_obj.return_value = (None, "signedvote")
        mock_decrypt.return_value = (None, json.dumps(self.deciphered_votes))
        mock_verifysig.return_value = (None, True)
        mock_fuse.return_value = "dummysecret"
        mock_router.return_value.fetch_votes.return_value = \
            (self.vote_objs, False)
        failed_votes = [{"failed": 1}, {"failed": 2}]
        mock_router.return_value.sign_votes.return_value = \
            {"invalid": 0, "failed": failed_votes}
        results_obj = ElectionResults(election_id=self.election_id)
        results_obj.save()
        utils.async_scrutinize(results_obj, self.received_data)
        results = json.loads(results_obj.election_results)
        # Results were properly published
        self.assertEquals(
            results_obj.result_status_choices,
            ElectionResults.ResultStatus.PUBLISHED
        )
        # All votes were counted
        self.assertEquals(results['total_votes'], len(self.vote_objs))
        # Two votes marked as invalid
        self.assertEquals(results['invalid_format'], len(failed_votes))
        # Valid votes except the failed ones
        self.assertEquals(
            results['total_valid'], len(self.vote_objs) - len(failed_votes)
        )

    @patch("core.utils.pkiutils.signObject")
    @patch("core.utils.pkiutils.decryptWithPrivateKey")
    @patch("core.utils.eccblind.verifySignature")
    @patch("core.utils.routers.ASrouter")
    @patch("core.utils.fuse_shared_keys")
    def test_async_scrutinize_invalid_format(self, mock_fuse, mock_router,
                                             mock_verifysig, mock_decrypt,
                                             mock_sign_obj):
        mock_sign_obj.return_value = (None, "signedvote")
        # Invalid JSON vote (missing 'id' attribute)
        invalid_vote = [self.create_vote(candidates=[{"vote": 1}])]
        mock_decrypt.return_value = (None, json.dumps(invalid_vote))
        mock_verifysig.return_value = (None, True)
        mock_fuse.return_value = "dummysecret"
        mock_router.return_value.fetch_votes.return_value = \
            ([self.create_vote_obj(id=1)], False)
        mock_router.return_value.sign_votes.return_value = \
            {"invalid": 0, "failed": []}
        results_obj = ElectionResults(election_id=self.election_id)
        results_obj.save()
        utils.async_scrutinize(results_obj, self.received_data)
        results = json.loads(results_obj.election_results)
        # Results were properly published
        self.assertEquals(
            results_obj.result_status_choices,
            ElectionResults.ResultStatus.PUBLISHED
        )
        # All votes were counted
        self.assertEquals(results['total_votes'], len(invalid_vote))
        # Two votes marked as invalid
        self.assertEquals(results['invalid_format'], len(invalid_vote))

    @patch("core.utils.pkiutils.signObject")
    @patch("core.utils.pkiutils.decryptWithPrivateKey")
    @patch("core.utils.eccblind.verifySignature")
    @patch("core.utils.routers.ASrouter")
    @patch("core.utils.fuse_shared_keys")
    def test_async_scrutinize_success(self, mock_fuse, mock_router,
                                      mock_verifysig, mock_decrypt,
                                      mock_sign_obj):
        mock_sign_obj.return_value = (None, "signedvote")
        mock_decrypt.return_value = (None, json.dumps(self.deciphered_votes))
        mock_verifysig.return_value = (None, True)
        mock_fuse.return_value = "dummysecret"
        mock_router.return_value.fetch_votes.return_value = \
            (self.vote_objs, False)
        mock_router.return_value.sign_votes.return_value = \
            {"invalid": 0, "failed": []}
        results_obj = ElectionResults(election_id=self.election_id)
        results_obj.save()
        utils.async_scrutinize(results_obj, self.received_data)
        results = json.loads(results_obj.election_results)
        # Results were properly published
        self.assertEquals(
            results_obj.result_status_choices,
            ElectionResults.ResultStatus.PUBLISHED
        )
        # All votes were counted
        self.assertEquals(results['total_votes'], len(self.vote_objs))
        # All votes marked as valid
        self.assertEquals(results['total_valid'], len(self.vote_objs))
        # Vote arithmetic
